/*****************************\
*                             *
*  Implements menu callbacks  *
*                             *
\*****************************/

#include <gtk/gtkmain.h>
#include "main.h"
#include "messagebox.h"
#include "aligndialog.h"
#include "rulerdialog.h"

TCADSheet *
TCADMain::GetActiveSheet()
{
  TCADSheet *Sheet;

	DTP("TCADMain::GetActiveSheet()\n");

  // Find if any sheet has focus
  Sheet = FirstSheet;
  while(Sheet) {
    if (Sheet->HasFocus()) break;
    Sheet = (TCADSheet *)Sheet->Next;
  }

  if (ActiveView == NULL) return(NULL);
  if (Sheet == NULL) Sheet = ActiveView->Sheet; // Use last active sheet

  return(Sheet);
}

void
TCADMain::Menu_New()
{
  TCADSheet *NewSheet;

	DTP("TCADMain::Menu_New()\n");

  // Add new sheet
  NewSheet = new TCADSheet(TRUE);
  AddSheet(NewSheet);
}

void
TCADMain::Menu_Open()
{
	DTP("TCADMain::Menu_Open()\n");

  FileOper = FO_LOAD;

  // Hide file selection dialog
  gtk_widget_hide(GTK_WIDGET(FileSelection));

  // Set title
  gtk_window_set_title(&FileSelection->window,"Load sheet");

  // Show it
  gtk_widget_show(GTK_WIDGET(FileSelection));
}

gboolean
TCADMain::Menu_SaveAll()
{
	TCADSheet *Sheet = FirstSheet;

	DTP("TCADMain::Menu_SaveAll()\n");

	while(Sheet) {
		Save(Sheet);
		Sheet = (TCADSheet *)Sheet->Next;
	}
	
  return(FALSE);
}

void
TCADMain::Menu_Exit()
{
  TCADSheet *Next;

	DTP("TCADMain::Menu_Exit()\n");

	ShuttingDown=true;

  // Close all sheets
  while(FirstSheet) {
    if (!CloseSheet(FirstSheet)) return;  // Dialog shown
  }

  // Exit
  gtk_main_quit();
}

void
TCADMain::Menu_NextWindow()
{
	DTP("TCADMain::Menu_NextWindow()\n");

/*
  if (ActiveSheet != NULL) {
    ActiveSheet = (TCADSheet *)ActiveSheet->Next;
    if (ActiveSheet == NULL) ActiveSheet = FirstSheet;
    ActiveSheet->BringToFront();
  }
*/
}

void
TCADMain::Menu_PrevWindow()
{
	DTP("TCADMain::Menu_PrevWindow()\n");

/*
  if (ActiveSheet != NULL) {
    ActiveSheet = (TCADSheet *)ActiveSheet->Prev;
    if (ActiveSheet == NULL) ActiveSheet = LastSheet;
    ActiveSheet->BringToFront();
  }
*/
}

void
TCADMain::Menu_CloseWindow()
{
  TCADSheet *Sheet;

	DTP("TCADMain::Menu_CloseWindow()\n");

  Sheet = GetActiveSheet();

  if (Sheet == NULL) return;

  // Really close
  if (Sheet->RemoveView(ActiveView)) CloseSheet(Sheet);
}

void
TCADMain::Menu_About()
{
  char Msg[1024];

	DTP("TCADMain::Menu_About()\n");

  sprintf(Msg,"%s %s\n\nCopyright 1998,9\nMichal Kra\nlemming@k332.feld.cvut.cz\n\nDistributed under GPL",PACKAGE,VERSION);
  MessageBox(Msg,"About");
}

void TCADMain::Menu_Export_Latex(TCADSheet *Sheet)
{
  FileOper = FO_EXPORT_LATEX;

  SaveSheet = Sheet;
  if (SaveSheet == NULL) return;

  // Hide file selection dialog
  gtk_widget_hide(GTK_WIDGET(FileSelection));

  // Set title
  gtk_window_set_title(&FileSelection->window,"Export as LaTeX picture");

  // Show it
  gtk_widget_show(GTK_WIDGET(FileSelection));

//  gtk_main();
}

void
TCADMain::Menu_Export_FIG(TCADSheet *Sheet)
{
	DTP("TCADMain::Menu_Export_FIG()\n");

  FileOper = FO_EXPORT_FIG;

  SaveSheet = Sheet;
  if (SaveSheet == NULL) return;

  // Hide file selection dialog
  gtk_widget_hide(GTK_WIDGET(FileSelection));

  // Set title
  gtk_window_set_title(&FileSelection->window,"Export as XFIG picture");

  // Show it
  gtk_widget_show(GTK_WIDGET(FileSelection));

//  gtk_main();
}

void
TCADMain::Menu_Export_EPS(TCADSheet *Sheet)
{
	DTP("TCADMain::Menu_Export_EPS()\n");

  FileOper = FO_EXPORT_EPS;

  SaveSheet = Sheet;
  if (SaveSheet == NULL) return;

  // Hide file selection dialog
  gtk_widget_hide(GTK_WIDGET(FileSelection));

  // Set title
  gtk_window_set_title(&FileSelection->window,"Export as Encapsulated PostScript");

  // Show it
  gtk_widget_show(GTK_WIDGET(FileSelection));
}

