/**********************\
*                      *
*  Read/write streams  *
*                      *
\**********************/

#ifndef GTKFIG_STREAMS_H
#define GTKFIG_STREAMS_H

#include <stdio.h>
#include <stdlib.h>
#include <glib.h>

/*
 * Memory - for memory streams
 */
class TCADMemory {
protected:
	unsigned int Size,Step;
public:
	TCADMemory(unsigned int StartSize=20480,unsigned int AllocStep=10240) : Size(StartSize),Step(AllocStep) { Memory = malloc(StartSize); }
	~TCADMemory() { free(Memory); }

	void *Memory;

	// Enlarges chunk if necessary
	void EnsureSize(int NeededSize);
};

/*
 * Streams for saving
 */
class TCADSaveStream {
public:
	TCADSaveStream() : SaveOnlySelected(false) {}

	gboolean SaveOnlySelected;
  virtual void SaveInt(char *Name,int Data) = 0;
  virtual void SaveString(char *Name,char *Data) = 0;
  virtual void SaveObjectName(char *Name) = 0;
  virtual void SaveEndObject() = 0;
};

class TCADFileSaveStream : public TCADSaveStream {
protected:
  static const unsigned int MAXDEPTH=100;
  FILE *Fl;
  int Depth;
  char Spaces[MAXDEPTH+3];
public:
  TCADFileSaveStream(char *FileName,char &Success);
  virtual ~TCADFileSaveStream();
	
  virtual void SaveInt(char *Name,int Data);
  virtual void SaveString(char *Name,char *Data);
  virtual void SaveObjectName(char *Name);
  virtual void SaveEndObject();
};

class TCADMemorySaveStream : public TCADSaveStream {
protected:
	TCADMemory *mem;
	int off;
public:
	TCADMemorySaveStream(TCADMemory &Memory) : mem(&Memory),off(0) {} 

  virtual void SaveInt(char *Name,int Data);
  virtual void SaveString(char *Name,char *Data);
  virtual void SaveObjectName(char *Name);
  virtual void SaveEndObject();

	// For debugging only
	void SaveToFile(char *FileName);
};

/*
 * Streams for loading
 */
class TCADLoadStream {
protected:
  static const unsigned int BUFSIZE=1024;
  static const unsigned int MAXIDLEN=64;

  char ID[MAXIDLEN+3];
  char Buffer[BUFSIZE+3];

  char WasLF;

  void ReadROL();  // Reads rest of line

	virtual int NextChar()=0;

	GHashTable *UniqTable;
public:
	TCADLoadStream():UniqTable(NULL),RenumberUniq(false),KeepChanges(false) {}
	~TCADLoadStream() { if (UniqTable) g_hash_table_destroy(UniqTable); }

	gboolean RenumberUniq,KeepChanges;
	
  virtual char *ReadID();
  virtual char *LoadString();
  virtual int LoadInt();

  // If ID was value, skips value, if ID was object, skips whole object
  virtual void Skip();

	/* Functions for renumbering Uniq (when pasting) */
	void UniqTranslated(unsigned long Old,unsigned long New);
	unsigned long TranslateUniq(unsigned long Old);
};
 
class TCADFileLoadStream : public TCADLoadStream {
protected:
  FILE *Fl;

	virtual int NextChar() { return((Fl==NULL)?-1:fgetc(Fl)); }
public:
  TCADFileLoadStream(char *FileName,char &Success);
  ~TCADFileLoadStream();
};

class TCADMemoryLoadStream : public TCADLoadStream {
protected:
  TCADMemory *mem;
	unsigned char *Ptr;

	virtual int NextChar() { return((*Ptr==0)?-1:(*Ptr++)); }
public:
  TCADMemoryLoadStream(TCADMemory &Memory) : mem(&Memory) { Ptr=(unsigned char *)(mem->Memory); }
};


#endif

