/*****************\
*                 *
*  View of sheet  *
*                 *
\*****************/

#include <gtk/gtk.h>
#include <math.h>
#include <memory.h>

#include "view.h"
#include "sheet.h"
#include "main.h"
#include "line.h"
#include "viewmenu.h"

static char FocusDash[2] = {3, 3};

/*
 * Menu
 */

#define MENU_NITEMS 46
static GtkItemFactoryEntry menu_items[MENU_NITEMS] =
{
  {"/Popup menu",NULL,NULL,0,"<Title>"}, 
  {"/File", NULL, NULL, 0, "<Branch>"},
  {"/File/New", "<control>N", VMA_New, 0, NULL},
  {"/File/Open", "<control>O", VMA_Open, 0, NULL},
  {"/File/Save", "<control>S", VMA_Save, 0, NULL},
  {"/File/Save as ...", NULL, VMA_SaveAs, 0, NULL},
  {"/File/Save all", NULL, VMA_SaveAll, 0, NULL},
  {"/File/Export", NULL, NULL, 0, "<Branch>"},
  {"/File/Export/LaTeX", NULL, VMA_Export_Latex, 0, NULL},
  {"/File/Export/EPS", NULL, VMA_Export_EPS, 0, NULL},
  {"/File/Export/FIG", NULL, VMA_Export_FIG, 0, NULL},
  {"/File/<separator>", NULL, NULL, 0, "<Separator>"},
  {"/File/Close", "<control>W", VMA_Close, 0, NULL},
  {"/File/Exit", "<control>Q", VMA_Exit, 0, NULL},
	{"/Edit",NULL, NULL, 0, "<Branch>"},
  {"/Edit/Undo", "<control>Z", VMA_Undo, 0, NULL},
  {"/Edit/Redo", "<control>Y", VMA_Redo, 0, NULL},
  {"/Edit/<separator>", NULL, NULL, 0, "<Separator>"},
  {"/Edit/Delete", "delete", VMA_Delete, 0, NULL},
  {"/Edit/Paste", "<control>V", VMA_Paste, 0, NULL},
  {"/Edit/Copy", "<control>C", VMA_Copy, 0, NULL},
  {"/Edit/Cut", "<control>X", VMA_Cut, 0, NULL},
  {"/Edit/<separator>", NULL, NULL, 0, "<Separator>"},
  {"/Edit/Select all", "<control>A", VMA_SelectAll, 0, NULL},
  {"/View", NULL, NULL, 0, "<Branch>"},
  {"/View/New", NULL, VMA_NewView, 0, NULL},
  {"/View/Close", "<control>0", VMA_CloseView, 0, NULL},
  {"/View/<separator>", NULL, NULL, 0, "<Separator>"},
  {"/View/Zoom in", "equal", VMA_ZoomIn, 0, NULL},
  {"/View/Zoom out", "minus", VMA_ZoomOut, 0, NULL},
  {"/View/Zoom 1:1", "1", VMA_ZoomReset, 0, NULL},
  {"/Objects", NULL, NULL, 0, "<Branch>"},
  {"/Objects/Align", NULL, VMA_Align, 0, NULL},
  {"/Objects/Edit text", "<control>E", VMA_EditText, 0, NULL},
  {"/Objects/Send to back", NULL, VMA_SendToBack, 0, NULL},
  {"/Objects/Bring to front", NULL, VMA_BringToFront, 0, NULL},
  {"/Objects/<separator>", NULL, NULL, 0, "<Separator>"},
  {"/Objects/Text properties", NULL, VMA_TextProperties, 0, NULL},
  {"/Objects/Line properties", NULL, VMA_LineProperties, 0, NULL},
  {"/Objects/Arrow properties", NULL, VMA_ArrowProperties, 0, NULL},
  {"/Grid", NULL, NULL, 0, "<Branch>"},
  {"/Grid/Grid options", NULL, VMA_GridOptions, 0, NULL},
  {"/Grid/Use grid", NULL, VMA_UseGrid, 0, "<CheckItem>"},
  {"/Grid/Show grid", NULL, VMA_ShowGrid, 0, "<CheckItem>"},
  {"/Sheet size", NULL, VMA_Size, 0, NULL},
  {"/Rulers", NULL, VMA_Rulers, 0, NULL},
};

// Alignment macro
#define ALIGN(x,gran,Force) \
if ((Sheet->GetUseGrid()) || (Force)) { \
  int _T = x % gran; \
  if (_T < (gran/2)) x -= _T; /* Round down */ \
  else x += (gran-_T); /* Round up  */\
}

/*
 * Event functions
 */

static gboolean
window_delete_event (GtkWidget       *widget,
                     GdkEvent        *event,
                     gpointer         user_data)
{
  TCADView *View;
  widget = widget; event = event;

	DTP("window_delete_event()\n");

  if (gtk_grab_get_current() != NULL) return(true);  // Modal dialog is showing

  View = (TCADView *)user_data;
	if (View->Sheet->Closing) return(false);  // Work already being done
	
  if (View->Sheet->RemoveView(View)) {
		System->CloseSheet(View->Sheet);
		return(false);
	}
	else return(true);
}

static gboolean
visibility_notify_event (GtkWidget          *widget,
                         GdkEventVisibility *event,
                         gpointer            user_data)
{
  TCADView *View;
  widget = widget;
  event = event;

//	DTP("visibility_notify_event()\n");

//  printf("visibility_notify_event()\n");
  View = (TCADView *)user_data;
  View->SetVisibilityState(event->state);
  return TRUE;
}

static gboolean
focus_in_event (GtkWidget     *widget,
                GdkEventFocus *event,
                gpointer       user_data)
{
  TCADView *View;
  widget = widget;
  event = event;

//	DTP("focus_in_event()\n");

  View = (TCADView *)user_data;
  View->GotFocus();
  return TRUE;
}

static gboolean
fixed_expose_event (GtkWidget       *widget,
                    GdkEventExpose  *event,
                    gpointer         user_data)
{
  TCADView *View;
	int MagA,MagB;
  widget = widget;

//	DTP("fixed_expose_event()\n");

  View = (TCADView *)user_data;
	View->GetMag(MagA,MagB);
//	printf("Expose: X=%d, Y=%d, W=%d, H=%d\n",event->area.x,event->area.y,event->area.width,event->area.height);
	event->area.x = event->area.x*MagB/MagA;
	event->area.y = event->area.y*MagB/MagA;
	event->area.width = event->area.width*MagB/MagA;
	event->area.height = event->area.height*MagB/MagA;
	View->Sheet->DrawRect(&event->area);
  return TRUE;
}

static void
fixed_draw  (GtkWidget       *widget,
             GdkRectangle    *area,
             gpointer         user_data)
{
  TCADView *View;
	int MagA,MagB;
  widget = widget;

//	DTP("fixed_draw()\n");

  View = (TCADView *)user_data;
	View->GetMag(MagA,MagB);
//	printf("Expose: X=%d, Y=%d, W=%d, H=%d\n",area->x,area->y,area->width,area->height);
	area->x = area->x*MagB/MagA;
	area->y = area->y*MagB/MagA;
	area->width = area->width*MagB/MagA;
	area->height = area->height*MagB/MagA;
  View->Sheet->DrawRect(area);
}

static gboolean
fixed_button_press_event (GtkWidget      *widget,
                          GdkEventButton *event,
                          gpointer       user_data)
{
  TCADView *View;
  widget = widget;

	DTP("fixed_button_press_event()\n");

  View = (TCADView *)user_data;
//  printf("event->state = %x\n",event->state);
//  Sheet = (TCADSheet *)gtk_object_get_data((GtkObject *)widget,"Sheet");
  View->ButtonPressed((int)event->x,(int)event->y,event->state,event->button);

  return TRUE;
}

static gboolean
fixed_button_release_event (GtkWidget      *widget,
                            GdkEventButton *event,
                            gpointer       user_data)
{
  TCADView *View;
  widget = widget;

	DTP("fixed_button_release_event()\n");

  if (event->button != 1) return FALSE;

  View = (TCADView *)user_data;
//  Sheet = (TCADSheet *)gtk_object_get_data((GtkObject *)widget,"Sheet");
  View->ButtonReleased();

  return TRUE;
}
static gboolean
fixed_motion_notify_event (GtkWidget      *widget,
                           GdkEventMotion *event,
                           gpointer       user_data)
{
  TCADView *View;
  widget = widget;

//	DTP("fixed_motion_notify_event()\n");

  View = (TCADView *)user_data;
//  Sheet = (TCADSheet *)gtk_object_get_data((GtkObject *)widget,"Sheet");
  View->MouseMoved((int)event->x,(int)event->y);

  return FALSE;
}

/******************************************************************************/

/*
 * Class implementation
 */

#define STD_MASK GDK_EXPOSURE_MASK+GDK_BUTTON_PRESS_MASK+GDK_BUTTON_RELEASE_MASK+GDK_VISIBILITY_NOTIFY_MASK

TCADView::TCADView(TCADSheet *Sheet,char Visible)
{
  int nmenu_items = sizeof(menu_items) / sizeof(menu_items[0]);
  GtkTable *Table;
	GtkWidget *Label1,*Label2;

	DTP("TCADView::TCADView()\n");

  AccelGroup = gtk_accel_group_new();
  MagA = 1;
  MagB = 1;
  GC = NULL; // No GC for now
  ButtonDown = 0;
  Mode = TM_NOTHING;
  ResetOnUp = 0;
  Fixed = 0;
	bzero(&FocusRect,sizeof(FocusRect));

  this->Sheet = Sheet;

  VisibilityState = GDK_VISIBILITY_UNOBSCURED;

  VRulerSettings.Visible = TRUE;
  VRulerSettings.Metric = GTK_CENTIMETERS;
  HRulerSettings.Visible = TRUE;
  HRulerSettings.Metric = GTK_CENTIMETERS;

  /* Create window */
  Window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_signal_connect ((GtkObject *)(Window), "delete_event",
                      GTK_SIGNAL_FUNC (window_delete_event),
                      this);
  gtk_signal_connect ((GtkObject *)(Window), "visibility_notify_event",
                      GTK_SIGNAL_FUNC (visibility_notify_event),
                      this);
  gtk_signal_connect ((GtkObject *)(Window), "focus_in_event",
                      GTK_SIGNAL_FUNC (focus_in_event),
                      this);
  gtk_object_set_data((GtkObject *)(Window), "View", this);
  gtk_window_position (GTK_WINDOW (Window), GTK_WIN_POS_NONE);
  gtk_window_set_policy (GTK_WINDOW (Window), TRUE, TRUE, TRUE);
  gtk_widget_set_usize (Window, 640, 480);

  // Create scrolled window
  ScrolledWindow = GTK_SCROLLED_WINDOW(gtk_scrolled_window_new(NULL, NULL));
  gtk_object_set_data ((GtkObject *) (Window), "ScrolledWindow", ScrolledWindow);
  gtk_object_set_data((GtkObject *)(ScrolledWindow), "Sheet", this);
  gtk_widget_show (GTK_WIDGET(ScrolledWindow));
  gtk_container_add (GTK_CONTAINER (Window), GTK_WIDGET(ScrolledWindow));

  // Create table
  Table = GTK_TABLE(gtk_table_new(3,3,FALSE));
  gtk_table_set_col_spacing(Table,0,1);
  gtk_table_set_row_spacing(Table,0,1);
  gtk_container_border_width(GTK_CONTAINER(Table),2);
  gtk_scrolled_window_add_with_viewport(ScrolledWindow,GTK_WIDGET(Table));

  /* Create table and rulers with frame */
  Frame = GTK_FRAME(gtk_frame_new(NULL));
  gtk_frame_set_shadow_type(Frame,GTK_SHADOW_OUT);
  VRuler = GTK_VRULER(gtk_vruler_new());
  HRuler = GTK_HRULER(gtk_hruler_new());
  gtk_ruler_set_range (GTK_RULER(HRuler), 0, Sheet->EncapRect.width, 0, Sheet->EncapRect.width);
  gtk_widget_set_usize (GTK_WIDGET(HRuler), Sheet->EncapRect.width, -1);
  gtk_ruler_set_range (GTK_RULER(VRuler), 0, Sheet->EncapRect.height, 0, Sheet->EncapRect.height);
  gtk_widget_set_usize (GTK_WIDGET(VRuler), -1, Sheet->EncapRect.height);

  // Create fixed
  Fixed = gtk_fixed_new();
  gtk_object_set_data((GtkObject *)(Window), "Fixed", Fixed);
  gtk_object_set_data((GtkObject *)(Fixed), "Sheet", this);
  gtk_widget_set_usize (Fixed, Sheet->EncapRect.width, Sheet->EncapRect.height);
  gtk_signal_connect ((GtkObject *) (Fixed), "draw",
                      GTK_SIGNAL_FUNC (fixed_draw),
                      this);
  gtk_signal_connect ((GtkObject *) (Fixed), "expose_event",
                      GTK_SIGNAL_FUNC (fixed_expose_event),
                      this);
  gtk_signal_connect ((GtkObject *) (Fixed), "button_press_event",
                      GTK_SIGNAL_FUNC (fixed_button_press_event),
                      this);
  gtk_signal_connect ((GtkObject *) (Fixed), "button_release_event",
                      GTK_SIGNAL_FUNC (fixed_button_release_event),
                      this);
  gtk_signal_connect ((GtkObject *) (Fixed), "motion_notify_event",
                      GTK_SIGNAL_FUNC (fixed_motion_notify_event),
                      this);
  gtk_widget_set_events(Fixed,STD_MASK+GDK_POINTER_MOTION_MASK);

	Label1 = gtk_fixed_new ();
	gtk_widget_show(Label1);
	Label2 = gtk_fixed_new ();
	gtk_widget_show(Label2);

  // Insert widgets to table
  gtk_table_attach(Table,GTK_WIDGET(Frame), 0, 1, 0, 1, GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach(Table,GTK_WIDGET(HRuler), 1, 2, 0, 1, (GtkAttachOptions)0, (GtkAttachOptions)0, 0, 0);
  gtk_table_attach (Table, GTK_WIDGET(VRuler), 0, 1, 1, 2,
   (GtkAttachOptions)0, (GtkAttachOptions)0, 0, 0);
  gtk_table_attach (Table, GTK_WIDGET(Fixed), 1, 3, 1, 3,
   ((GtkAttachOptions)(GTK_EXPAND + GTK_SHRINK + GTK_FILL)), ((GtkAttachOptions)(GTK_EXPAND + GTK_SHRINK + GTK_FILL)), 0, 0);
  gtk_widget_show(Fixed);
  gtk_widget_show(GTK_WIDGET(Table));
  ApplyRulerSettings();

  /* Create menu */
  Factory = gtk_item_factory_new(GTK_TYPE_MENU,"<Test>",AccelGroup);
  gtk_item_factory_create_items(Factory,MENU_NITEMS,menu_items,this);
	gtk_window_add_accel_group(GTK_WINDOW(Window), Factory->accel_group);
  gtk_table_attach (GTK_TABLE (Table), Label1, 0, 1, 2, 3,
                    (GtkAttachOptions) 0, (GtkAttachOptions)(GTK_EXPAND | GTK_FILL), 0, 0);
  gtk_table_attach (GTK_TABLE (Table), Label2, 2, 3, 0, 1,
                    (GtkAttachOptions)(GTK_EXPAND | GTK_FILL), (GtkAttachOptions) 0, 0, 0);

  // Get some widgets
	MIUseGrid = GTK_CHECK_MENU_ITEM(gtk_item_factory_get_widget(Factory,"/Grid/Use grid"));
	MIShowGrid = GTK_CHECK_MENU_ITEM(gtk_item_factory_get_widget(Factory,"/Grid/Show grid"));
	
  gtk_check_menu_item_set_show_toggle(MIShowGrid,TRUE);
  gtk_check_menu_item_set_show_toggle(MIUseGrid,TRUE);

	// Disable menu items
//	gtk_widget_set_sensitive(gtk_item_factory_get_widget(Factory,"/File/Export/EPS"),FALSE);

  RefreshCheckItems();
	
  // Add yourself
  Sheet->AddView(this);

  if (Visible) Show();
}

TCADView::~TCADView()
{
	DTP("TCADView::~TCADView()\n");

  if (GC != NULL) gdk_gc_destroy(GC);

/*
  gtk_accel_group_destroy(AccelGroup);
  gtk_item_factory_destroy(Factory);
*/

	gtk_widget_hide(Window);
	gtk_widget_destroy(Window);
}

void
TCADView::SetTitle(char *Title)
{
	DTP("TCADView::SetTitle()\n");

  gtk_window_set_title(GTK_WINDOW (Window), Title);
}

void
TCADView::GotFocus()
{
	DTP("TCADView::GotFocus()\n");

//  if (VisibilityState == GDK_VISIBILITY_FULLY_OBSCURED)
   System->ViewActivated(this);
}

gboolean
TCADView::HasFocus()
{
	DTP("TCADView::HasFocus()\n");

  return(GTK_WIDGET_HAS_FOCUS(Window));
}

void
TCADView::BringToFront()
{
	DTP("TCADView::BringToFront()\n");

  gdk_window_raise(Window->window);
}

void
TCADView::ApplyRulerSettings()
{
	DTP("TCADView::ApplyRulerSettings()\n");

  if (VRulerSettings.Visible) gtk_widget_show(GTK_WIDGET(VRuler));
  else gtk_widget_hide(GTK_WIDGET(VRuler));

  if (HRulerSettings.Visible) gtk_widget_show(GTK_WIDGET(HRuler));
  else gtk_widget_hide(GTK_WIDGET(HRuler));

  if (HRulerSettings.Visible && VRulerSettings.Visible) gtk_widget_show(GTK_WIDGET(Frame));
  else gtk_widget_hide(GTK_WIDGET(Frame));

  gtk_ruler_set_metric(GTK_RULER(VRuler),VRulerSettings.Metric);
  gtk_ruler_set_metric(GTK_RULER(HRuler),HRulerSettings.Metric);

	Sheet->Changed();
}

GdkDrawable *
TCADView::GetDrawable()
{
//	DTP("TCADView::GetDrawable()\n");

  if (Fixed == NULL) return(NULL);
  else return(Fixed->window);
}

GdkWindow *
TCADView::GetWindow()
{
	DTP("TCADView::GetWindow()\n");

  if (Fixed == NULL) return(NULL);
  else return(Fixed->window);
}

GtkWidget *
TCADView::GetGtkWindow()
{
	DTP("TCADView::GetGtkWindow()\n");

	return(Window);
}

void
TCADView::Show()
{
  GdkColormap *ColorMap;

	DTP("TCADView::Show()\n");

  gtk_widget_show(Window);

  if (GC == NULL) {
    GC = gdk_gc_new(Fixed->window);

    // Get colors
    ColorMap = gdk_window_get_colormap(Fixed->window);
    gdk_color_white(ColorMap,&WhiteColor);
    gdk_color_black(ColorMap,&BlackColor);

    BlueColor.pixel = 0;
    BlueColor.red = 0;
    BlueColor.green = 0;
    BlueColor.blue = 65535;
    gdk_colormap_alloc_color(ColorMap,&BlueColor,TRUE,TRUE);

		// Resize
		if ((WndX != 0) || (WndY != 0)) gdk_window_move(Window->window,WndX,WndY);

		// Refresh items
		RefreshCheckItems();
  }
}

void
TCADView::DrawFocusRect(int MagA,int MagB)
{
	DTP("TCADView::DrawFocusRect()\n");

  gdk_gc_set_foreground(GC,&BlackColor);
  gdk_gc_set_function(GC,GDK_INVERT);
	gdk_gc_set_line_attributes(GC,1,GDK_LINE_ON_OFF_DASH,GDK_CAP_NOT_LAST,GDK_JOIN_MITER);
  gdk_gc_set_dashes(GC,0,FocusDash,2);

//  printf("DrawFocusRect(%d,%d,%d,%d)\n",FocusRect.x,FocusRect.y,FocusRect.width,FocusRect.height);  

  gdk_draw_rectangle(Fixed->window,GC,0,FocusRect.x*MagA/MagB,FocusRect.y*MagA/MagB,FocusRect.width*MagA/MagB,FocusRect.height*MagA/MagB);

  gdk_gc_set_function(GC,GDK_COPY);
  gdk_gc_set_line_attributes(GC,1,GDK_LINE_SOLID,GDK_CAP_NOT_LAST,GDK_JOIN_MITER);
}

void
TCADView::ButtonPressed(int x,int y,int State,int Button)
{
  TCADLine *Line;
  int GridX,GridY;
  gboolean T;
  GdkRectangle RefreshRect;
  gboolean Shift;
  int x2,y2;

	DTP("TCADView::ButtonPressed()\n");

  Shift = State & GDK_SHIFT_MASK;

  if (Button == 3) {
    gdk_window_get_position(GTK_WIDGET(Window)->window,&x2,&y2);
    x += x2;
    y += y2;
    x += GTK_WIDGET(Fixed)->allocation.x;
    y += GTK_WIDGET(Fixed)->allocation.y;
		x -= (int)(GTK_RANGE(ScrolledWindow->hscrollbar)->old_value);
		y -= (int)(GTK_RANGE(ScrolledWindow->vscrollbar)->old_value);
    gtk_item_factory_popup(Factory,x,y,State,0);
    return;
  }


  if (Button != 1) return;

  Sheet->GetGridOptions(T,GridX,GridY);

  x = x*MagB/MagA;
  y = y*MagB/MagA;

  DownX = x;
  DownY = y;
	SGCorrX = 0;
	SGCorrY = 0;
  ResetOnUp = FALSE;
  ButtonDown = TRUE;

  System->ViewActivated(this);

  if (System->GetState() != ST_PICK) {
    /* We will create object */

    // Align x,y,DownX,DownY
    ALIGN(DownX,GridX,FALSE);
    ALIGN(DownY,GridY,FALSE);
    XCorr = x-DownX;
    YCorr = y-DownY;
    ALIGN(x,GridX,FALSE);
    ALIGN(y,GridY,FALSE);
    OldX = x;
    OldY = y;

    switch(System->GetState()) {
      case ST_LINE:
      case ST_ANGLELINE:
				Sheet->BlockChange++;
        Sheet->UnselectAll();
        Line = Sheet->CreateLine(System->GetState(),x,y);

        Mode = TM_DRAGHANDLE;
        DHType = HT_LINEEND;
        Line->HandleButtonPressed(0,x,y);
        MouseMoved(x*MagA/MagB,y*MagA/MagB);  // Eventually attach
        Line->HandleButtonReleased(&RefreshRect,this);
        Line->HandleButtonPressed(1,x,y);
        MouseMoved(x*MagA/MagB,y*MagA/MagB);  // Eventually attach
  
        // Redraw
        RefreshRect.x = x-HANDLE_SIZE/2;
        RefreshRect.y = y-HANDLE_SIZE/2;
        RefreshRect.width = HANDLE_SIZE+1;
        RefreshRect.height = HANDLE_SIZE+1;
        Sheet->DrawRect(&RefreshRect);
				Sheet->BlockChange--;
        ResetOnUp = 1;
        return;
      case ST_FRAME:
      case ST_TEXT:
      case ST_FRAMED_TEXT:
        Sheet->UnselectAll();  // Unselect objects
        Mode = TM_NEWRECT;
        ResetOnUp = TRUE;
        XCorr = 0;
        YCorr = 0;

        UpdateFocusRect(x,y);
        return;
      default: ;
    }
  }

	if (System->GetState() == ST_PICK) {
 	  XCorr = 0;
    YCorr = 0;
 	  OldX = 0;
   	OldY = 0;
	  switch(Sheet->ButtonPressed(x,y,Shift,this)) {
		case BPR_OBJECT:
	    Mode = TM_NOTHING;
    	return;
		case BPR_NOTHING:
	    Mode = TM_SELECTRECT;
  	  UpdateFocusRect(x,y);
    	return;
		case BPR_HANDLE:
			Mode = TM_DRAGHANDLE;
			return;
		}
	}
}

void
TCADView::ButtonReleased()
{
  int State;
  TCADObject *FirstSelected = Sheet->GetFirstSelected();
  GdkRectangle Rect1;

	DTP("TCADView::ButtonReleased()\n");

  ButtonDown = 0;

  State = System->GetState();
  if (ResetOnUp) System->ResetMode(FALSE);
  ResetOnUp = 0;

  switch(Mode) {
    case TM_SELECTRECT:
      // Clear clipping
      gdk_gc_set_clip_rectangle(GC,NULL);
      // Clear rectangle
      DrawFocusRect(MagA,MagB);
      Mode = TM_NOTHING;	// Reset mode
      
      Sheet->SelectRectangle(FocusRect.x,FocusRect.y,FocusRect.x+FocusRect.width,FocusRect.y+FocusRect.height);
      break;
    case TM_NEWRECT:
      // Clear clipping
      gdk_gc_set_clip_rectangle(GC,NULL);
      // Clear rectangle
      DrawFocusRect(MagA,MagB);

      // Reset our mode
      if (System->GetState() == ST_PICK) Mode = TM_NOTHING;

      if ((FocusRect.width < 4) && (FocusRect.height < 4)) return; // Too small - ignore

      // Enlarge FocusRect so that right/bottom boundary is also on grid
      FocusRect.width++;
      FocusRect.height++;

      Sheet->NewRectObject(State,&FocusRect);
      break;
    case TM_DRAGHANDLE:
      // Reset our mode
      if (System->GetState() == ST_PICK) Mode = TM_NOTHING;

      if (FirstSelected != NULL) {
				if (FirstSelected->HandleButtonReleased(&Rect1,this)) {
					Sheet->DrawRect(&Rect1);
					Sheet->Changed();
				}
			}
				
      break;
    case TM_DRAGGING:
      // Clear clipping
      gdk_gc_set_clip_rectangle(GC,NULL);

      // Clear rectangle
      DrawFocusRect(MagA,MagB);

      // Reset our mode
      if (System->GetState() == ST_PICK) Mode = TM_NOTHING;

      Sheet->MoveSelected(FocusRect.x-FocusX,FocusRect.y-FocusY);
      break;
  }
}

void
TCADView::MouseMoved(int x,int y)
{
  TAttach Attach;
  TCADObject *Obj;
  GdkRectangle Rect;
  char First;
  int T1,T2;
  int XOff,YOff;
  TCADObject *FirstSelected = Sheet->GetFirstSelected();
  gboolean Attached,T;
  int GridX,GridY;
	int tx,ty;
	
  Sheet->GetGridOptions(T,GridX,GridY);

//	DTP("TCADView::MouseMoved()\n");

  x = x*MagB/MagA;
  y = y*MagB/MagA;

  GTK_RULER(VRuler)->position = y;
  gtk_ruler_draw_pos(GTK_RULER(VRuler));
  GTK_RULER(HRuler)->position = x;
  gtk_ruler_draw_pos(GTK_RULER(HRuler));

  if (!ButtonDown) return;

  if (Mode != TM_SELECTRECT) {
    Attached = FALSE;

		if ((Mode == TM_DRAGGING) || (Mode == TM_NOTHING)) {
			x += XCorr;
			y += YCorr;
		}
 		ALIGN(x,GridX,FALSE);
		ALIGN(y,GridY,FALSE);
		
    if ((Mode == TM_DRAGHANDLE) && (DHType == HT_LINEEND)) {
      if (Sheet->CanAttach(x,y,&Attach)) {
        Attached = TRUE;
      }
      ((TCADLine *)(FirstSelected))->SetHandleAttachment(Attached,&Attach);
    }
  }

  if ((OldX == x) && (OldY == y)) return;  // Nothing changed
  OldX = x;
  OldY = y;

  if ((Mode == TM_NEWRECT) || (Mode == TM_SELECTRECT)) {
    // Clear clipping
    gdk_gc_set_clip_rectangle(GC,NULL);

    DrawFocusRect(MagA,MagB);
    UpdateFocusRect(x,y);
    return;
  }

  if (Mode == TM_DRAGHANDLE) {
    ALIGN(x,GridX,FALSE);
    ALIGN(y,GridY,FALSE);
    if (FirstSelected->HandleDragged(x,y,&Rect,this)) {
      // Redraw
      Sheet->DrawRect(&Rect);
    }
    return;
  }

  // Clear clipping
  gdk_gc_set_clip_rectangle(GC,NULL);

  // We start dragging
  if (Mode == TM_NOTHING) {
    // Get rectangle
    Obj = FirstSelected;
    First = 1;
    while(Obj) {
      if (First) FocusRect = Obj->EncapRect;
      else System->UnionRectangles(&FocusRect,&Obj->EncapRect);

      First = 0;
      Obj = Obj->NextSelected;

    }
		FocusRect.width--;
		FocusRect.height--;
    FocusX = FocusRect.x;
    FocusY = FocusRect.y;
		tx = FocusX;
		ty = FocusY;
		ALIGN(tx,GridX,FALSE);
		ALIGN(ty,GridY,FALSE);
		SGCorrX = (tx-FocusX);
		SGCorrY = (ty-FocusY);

    Mode = TM_DRAGGING;
  }
  else if (Mode == TM_DRAGGING) {
    // Mode is TM_DRAGGING
    DrawFocusRect(MagA,MagB);
		x -= SGCorrX;
		y -= SGCorrY;
  }

  /* Compute focus rect position */
  FocusRect.x = x;
  FocusRect.y = y;

  // Position inside sheet
  if (FocusRect.x < 0) FocusRect.x = 0;
  if (FocusRect.y < 0) FocusRect.y = 0;

  T1 = FocusRect.x+FocusRect.width-1;
  T2 = Sheet->EncapRect.x+Sheet->EncapRect.width-1;
  if (T1 > T2) FocusRect.x -= (T1-T2);

  T1 = FocusRect.y+FocusRect.height-1;
  T2 = Sheet->EncapRect.y+Sheet->EncapRect.height-1;
  if (T1 > T2) FocusRect.y -= (T1-T2);

  DrawFocusRect(MagA,MagB);
}

void
TCADView::HandleClicked(int x,int y,TCADHandle *Handle)
{
  TCADObject *FirstSelected = Sheet->GetFirstSelected();

	DTP("TCADView::HandleClicked()\n");

  // x and y are in normalized coordinates!
  XCorr = (Handle->x)-x;
  YCorr = (Handle->y)-y;
  Mode = TM_DRAGHANDLE;
  DHType = Handle->Type;
}

void
TCADView::UpdateFocusRect(int x,int y)
{
	DTP("TCADView::UpdateFocusRect()\n");

  FocusRect.x = MIN(DownX,x);
  FocusRect.y = MIN(DownY,y);
  FocusRect.width = ABS(DownX - x);
  FocusRect.height = ABS(DownY - y);

  // Clear clipping
  gdk_gc_set_clip_rectangle(GC,NULL);

  DrawFocusRect(MagA,MagB);
}

void
TCADView::Clear(GdkRectangle *Rect,int MagA,int MagB)
{
  int i,n,x,y,x2,y2,w,h;
  int GridX,GridY;
  gboolean T;
	double Ratio=((double)MagA)/MagB;
	GdkRectangle Rect2;

//	DTP("TCADView::Clear()\n");

  Sheet->GetGridOptions(T,GridX,GridY);

  // Set clipping
	Rect2.x = Rect->x*MagA/MagB;
	Rect2.y = Rect->y*MagA/MagB;
	Rect2.width = Rect->width*MagA/MagB;
	Rect2.height = Rect->height*MagA/MagB;
  gdk_gc_set_clip_rectangle(GC,&Rect2);

  // Clear area
  gdk_gc_set_background(GC,&WhiteColor);
  gdk_gc_set_foreground(GC,&WhiteColor);
  gdk_draw_rectangle(Fixed->window,GC,1,Rect2.x,Rect2.y,Rect2.width,Rect2.height);
  gdk_gc_set_foreground(GC,&BlackColor);

  // Draw grid
  if (Sheet->GetShowGrid()) {
		x = Rect->x;
    i = x % GridX;
    if (i > 0) x += (GridX - i);

		y = Rect->y;
    i = y % GridY;
    if (i > 0) y += (GridY - i);

    i = y;
    x2 = Rect->x+Rect->width;
    y2 = Rect->y+Rect->height;
    while(x<=x2) {
      y = i;
      while(y<=y2) {
				if (Fixed->window == NULL) {
					exit;
				}

        gdk_draw_point(Fixed->window,GC,(int)rint(x*Ratio),(int)rint(y*Ratio));
        y += GridY;
      }
      x += GridX;
    }
  }
}

void
TCADView::Draw(int MagA,int MagB)
{
  TCADObject *FirstSelected = Sheet->GetFirstSelected();

//	DTP("TCADView::Draw()\n");

  switch(Mode) {
    case TM_DRAGHANDLE:
      if (FirstSelected)
       if (FirstSelected->IsLine()) break;
    case TM_DRAGGING:
    case TM_NEWRECT:
    case TM_SELECTRECT:
      DrawFocusRect(MagA,MagB);
      break;
    default:
      break;
  }
}

void
TCADView::GetMag(int& A,int& B)
{
//	DTP("TCADView::GetMag()\n");

  A = MagA;
  B = MagB;
}

void
TCADView::SetMag(int A,int B)
{
	int W,H;
	int OldA,OldB;
	GtkAdjustment *ha,*va;
	float Mag1=(float)A/(float)B;
	float Mag2=(float)MagA/(float)MagB;
	float MagChange=Mag1/Mag2;

	DTP("TCADView::SetMag()\n");

	ha = gtk_scrolled_window_get_hadjustment(ScrolledWindow);
	va = gtk_scrolled_window_get_vadjustment(ScrolledWindow);
//	printf("ha->value1=%f, ha->lower1=%f, ha->upper1=%f\n",ha->value,ha->lower,ha->upper);

	OldA = MagA;
	OldB = MagB;
  MagA = A;
  MagB = B;

	Sheet->BlockChange++;
	Sheet->GetSheetSize(W,H);
	SetupCanvas(W,H);
	Sheet->BlockChange--;
	Sheet->Changed();
	
//	printf("ha->value2=%f, ha->lower2=%f, ha->upper2=%f\n",ha->value,ha->lower,ha->upper);


	gtk_adjustment_set_value(ha,ha->value*MagChange);
	gtk_adjustment_set_value(va,va->value*MagChange);

  Sheet->DrawRect(this,&Sheet->EncapRect,MagA,MagB);
}

void
TCADView::SetVisibilityState(GdkVisibilityState State)
{
	DTP("TCADView::SetVisibilityState()\n");

  VisibilityState = State;

  if ((State == GDK_VISIBILITY_FULLY_OBSCURED) && (GTK_WIDGET_HAS_FOCUS(Window)))
   System->ViewActivated(this);
}

void
TCADView::Hide()
{
	DTP("TCADView::Hide()\n");

  gtk_widget_hide(Window);
}

void
TCADView::RefreshCheckItems()
{
	DTP("TCADView::RefreshCheckItems()\n");

  gtk_check_menu_item_set_state(MIShowGrid,Sheet->GetShowGrid());
  gtk_check_menu_item_set_state(MIUseGrid,Sheet->GetUseGrid());
}

void
TCADView::SetRulersVisibility(int Visible)
{
	DTP("TCADView::SetRulersVisibility()\n");

	if (Visible) {
		gtk_widget_show(GTK_WIDGET(VRuler));
		gtk_widget_show(GTK_WIDGET(HRuler));
		gtk_widget_show(GTK_WIDGET(Frame));
	}
	else {
		gtk_widget_hide(GTK_WIDGET(VRuler));
		gtk_widget_hide(GTK_WIDGET(HRuler));
		gtk_widget_hide(GTK_WIDGET(Frame));
	}
}

void
TCADView::SetupWindow(int WndX,int WndY,int WndW,int WndH)
{
	DTP("TCADView::SetupWindow()\n");

	gtk_widget_set_usize(Window,WndW,WndH);

	// Store position to be set later
	this->WndX = WndX;
	this->WndY = WndY;
}

void
TCADView::SetupCanvas(int CanvasWidth,int CanvasHeight)
{
	DTP("TCADView::SetupCanvas()\n");

	double Ratio=((double)MagA)/MagB;
	gtk_widget_set_usize(Fixed, (int)rint(CanvasWidth*Ratio), (int)rint(CanvasHeight*Ratio));
  gtk_ruler_set_range (GTK_RULER(HRuler), 0, CanvasWidth, 0, CanvasWidth);
  gtk_widget_set_usize (GTK_WIDGET(HRuler), (int)rint(CanvasWidth*Ratio), -1);
  gtk_ruler_set_range (GTK_RULER(VRuler), 0, CanvasHeight, 0, CanvasHeight);
  gtk_widget_set_usize (GTK_WIDGET(VRuler), -1, (int)rint(CanvasHeight*Ratio));
}

void
TCADView::ZoomIn()
{
	int W,H;
	int MagA=this->MagA,MagB=this->MagB;

	DTP("TCADView::ZoomIn()\n");

	if (MagB > 1) {
		MagB >>= 1;
		MagA = 1;
	}
	else {
		MagB = 1;
		MagA <<= 1;
		if (MagA > 16) MagA = 16;
	}
	SetMag(MagA,MagB);
}

void
TCADView::ZoomOut()
{
	int MagA=this->MagA,MagB=this->MagB;

	DTP("TCADView::ZoomOut()\n");

	if (MagA > 1) {
		MagA >>= 1;
		MagB = 1;
	}
	else {
		MagA = 1;
		MagB <<= 1;
		if (MagB > 16) MagB = 16;
	}
	SetMag(MagA,MagB);
}
